﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;

namespace ErrorExample
{
    class Turtle
    {
        // 機体の幅
        public double PlatformWidth
        {
            get;
            set;
        }

        // 機体の高さ
        public double PlatformHeight
        {
            get;
            set;
        }

        // モーターが車輪を駆動する際のスピード。単位はメートル毎秒。
        // 簡単のため、タイヤが接地した状態で移動した距離とスリップした距離の
        // 合計から算出されていると仮定します。
        public double MotorSpeed
        {
            get;
            set;
        }

        // 左モーターの状態
        public MotorState LeftMotorState
        {
            get;
            set;
        }

        // 右モーターの状態
        public MotorState RightMotorState
        {
            get;
            set;
        }

        // タートルの現在位置
        public Point CurrentPosition
        {
            get;
            private set;
        }

        // タートルの現在の方向
        public double CurrentOrientation
        {
            get;
            private set;
        }

        // 特定の時間だけタートルを走行させます
        public void RunFor(double duration)
        {
            if (duration <= double.Epsilon)
            {
                throw new ArgumentException(
                    "0より長い時間を指定する必要があります",
                    "duration");
            }

            try
            {
                if (LeftMotorState == MotorState.Stopped &&
                    RightMotorState == MotorState.Stopped)
                {
                    // 完全に止まっている場合は何もしません
                    return;
                }
                // モーターが同じ方向に動いている場合には直進します
                if ((LeftMotorState == MotorState.Running &&
                    RightMotorState == MotorState.Running) ||
                    (LeftMotorState == MotorState.Reversed &&
                    RightMotorState == MotorState.Reversed))
                {
                    Drive(duration);
                }

                // モーターがそれぞれ反対方向に動いている場合には
                // 前進せずにタートルの中心を軸にしてただ回転します
                if ((LeftMotorState == MotorState.Running &&
                    RightMotorState == MotorState.Reversed) ||
                    (LeftMotorState == MotorState.Reversed &&
                    RightMotorState == MotorState.Running))
                {
                    Rotate(duration);
                }
            }
            catch (InvalidOperationException iox)
            {
                throw new Exception("タートルに何らかの問題が発生...", iox);
            }
            catch (Exception ex)
            {
                // この位置でログを出力
                Console.WriteLine("ログメッセージ：" + ex.Message);
                // 再スロー
                throw;
            }
            finally
            {
                Console.WriteLine("Turtleのfinallyブロック内");
            }
        }

        private void Rotate(double duration)
        {
            if (PlatformWidth <= 0.0)
            {
                throw new TurtleException(
                    "PlatformWidthは0.0より大きい値に初期化しなければいけません。");
            }

            // 旋回の総円周(circumference)
            double circum = Math.PI * PlatformWidth;

            // 総移動距離
            double d = duration * MotorSpeed;
            if (LeftMotorState == MotorState.Reversed)
            {
                // モーターが逆回転している場合は逆方向に回転します
                d *= -1.0;
            }

            // したがって一周のうちで回転した割合を以下の式で計算できます
            double proportionOfWholeCircle = d / circum;

            // 一回転が360度(2πラジアン)なので、
            // 車体の方向は以下のように変化したことになります
            CurrentOrientation =
                CurrentOrientation + (Math.PI * 2.0 * proportionOfWholeCircle);

            // return true; (このコードは不要なので削除できます)
        }


        private void Drive(double duration)
        {
            // 総移動距離
            double d = duration * MotorSpeed;
            if (LeftMotorState == MotorState.Reversed)
            {
                // モーターが逆回転している場合は逆方向に移動します
                d *= -1.0;
            }

            // 三角関数を使用してx,y座標の変化を計算します
            double deltaX = d * Math.Sin(CurrentOrientation);
            double deltaY = d * Math.Cos(CurrentOrientation);

            // 最後に位置を更新します
            CurrentPosition =
                new Point(CurrentPosition.X + deltaX, CurrentPosition.Y + deltaY);
        }
    }

    // モーターの現在の状態
    enum MotorState
    {
        Stopped,
        Running,
        Reversed
    }
}
