﻿using System;
using System.Windows;

namespace ErrorExample
{
    // 例6-5 範囲チェック
    class Turtle
    {
        // 機体の幅は1.0以上、10.0以下でなければならず、
        // 範囲外の値は範囲内の値に強制的に調整されます
        private double platformWidth;
        public double PlatformWidth
        {
            get { return platformWidth; }
            set
            {
                platformWidth = value;
                EnsurePlatformSize();
            }
        }

        // 機体の高さは1.0以上、10.0以下でなければならず、
        // 範囲外の値は範囲内の値に強制的に調整されます
        private double platformHeight;
        public double PlatformHeight
        {
            get { return platformHeight; }
            set
            {
                platformHeight = value;
                EnsurePlatformSize();
            }
        }

        // 新しいコンストラクタでは機体のサイズを適切に初期化します
        public Turtle()
        {
            EnsurePlatformSize();
        }

        // クラスの契約の更新によって新しく文書化された制約は
        // このメソッドによって保証されます
        private void EnsurePlatformSize()
        {
            if (PlatformWidth < 1.0)
            {
                PlatformWidth = 1.0;
            }
            if (PlatformWidth > 10.0)
            {
                PlatformWidth = 10.0;
            }
            if (PlatformHeight < 1.0)
            {
                PlatformHeight = 1.0;
            }
            if (PlatformHeight > 10.0)
            {
                PlatformHeight = 10.0;
            }
        }

        // モーターが車輪を駆動する際のスピード。単位はメートル毎秒。
        // 簡単のため、タイヤが接地した状態で移動した距離とスリップした距離の
        // 合計から算出されていると仮定します。
        public double MotorSpeed
        {
            get;
            set;
        }

        // 左モーターの状態
        public MotorState LeftMotorState
        {
            get;
            set;
        }

        // 右モーターの状態
        public MotorState RightMotorState
        {
            get;
            set;
        }

        // タートルの現在位置
        public Point CurrentPosition
        {
            get;
            private set;
        }

        // タートルの現在の方向
        public double CurrentOrientation
        {
            get;
            private set;
        }

        // 指定した時間だけタートルを走らせる
        public void RunFor(double duration)
        {
            if (LeftMotorState == MotorState.Stopped &&
            RightMotorState == MotorState.Stopped)
            {
                // 完全に止まっている場合は何もしません
                return;
            }

            // モーターが同じ方向に動いている場合には直進します
            if ((LeftMotorState == MotorState.Running &&
            RightMotorState == MotorState.Running) ||
            (LeftMotorState == MotorState.Reversed &&
                RightMotorState == MotorState.Reversed))
            {
                Drive(duration);
                return;
            }

            // モーターがそれぞれ反対方向に動いている場合には
            // 前進せずにタートルの中心を軸にしてただ回転します
            if ((LeftMotorState == MotorState.Running &&
            RightMotorState == MotorState.Reversed) ||
            (LeftMotorState == MotorState.Reversed &&
            RightMotorState == MotorState.Running))
            {
                Rotate(duration);
                return;
            }
        }

        private void Rotate(double duration)
        {
            // 旋回の総円周(circumference)
            double circum = Math.PI * PlatformWidth;

            // 総移動距離
            double d = duration * MotorSpeed;
            if (LeftMotorState == MotorState.Reversed)
            {
                // モーターが逆回転している場合は逆方向に回転します
                d *= -1.0;
            }

            // したがって一周のうちで回転した割合を以下の式で計算できます
            double proportionOfWholeCircle = d / circum;

            // 一回転が360度(2πラジアン)なので、
            // 車体の方向は以下のように変化したことになります
            CurrentOrientation =
                CurrentOrientation + (Math.PI * 2.0 * proportionOfWholeCircle);
        }

        private void Drive(double duration)
        {
            // 総移動距離
            double d = duration * MotorSpeed;
            if (LeftMotorState == MotorState.Reversed)
            {
                // モーターが逆回転している場合は逆方向に移動します
                d *= -1.0;
            }

            // 三角関数を使用してx,y座標の変化を計算します
            double deltaX = d * Math.Sin(CurrentOrientation);
            double deltaY = d * Math.Cos(CurrentOrientation);

            // 最後に位置を更新します
            CurrentPosition =
                new Point(CurrentPosition.X + deltaX, CurrentPosition.Y + deltaY);
        }
    }

    // モーターの現在の状態
    enum MotorState
    {
        Stopped,
        Running,
        Reversed
    }
}
