﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Security.AccessControl;
using System.Security.Principal;
using System.Text;

namespace DuplicateFinder
{
    class Program
    {
        static void Main(string[] args)
        {
            bool recurseIntoSubdirectories = false;

            if (args.Length < 1)
            {
                ShowUsage();
                return;
            }

            int firstDirectoryIndex = 0;
            IEnumerable<string> directoriesToSearch = null;
            bool testDirectoriesMade = false;

            // テストモードかどうかチェック
            if (args.Length == 1 && args[0] == "/test")
            {
                directoriesToSearch = MakeTestDirectories();
                testDirectoriesMade = true;
                recurseIntoSubdirectories = true;
            }
            else
            {
                if (args.Length > 1)
                {
                    // 再帰処理が必要かどうかチェック
                    if (args[0] == "/sub")
                    {
                        if (args.Length < 2)
                        {
                            ShowUsage();
                            return;
                        }
                        recurseIntoSubdirectories = true;
                        firstDirectoryIndex = 1;
                    }
                }

                // コマンドラインからディレクトリのリストを取得
                directoriesToSearch = args.Skip(firstDirectoryIndex);
            }

            try
            {
                List<FileNameGroup> filesGroupedByName =
                    InspectDirectories(recurseIntoSubdirectories, directoriesToSearch);

                DisplayMatches(filesGroupedByName);

                Console.ReadKey();
            }
            catch (PathTooLongException ptlx)
            {
                Console.WriteLine("指定されたパスが長すぎます。");
                Console.WriteLine(ptlx.Message);
            }
            catch (DirectoryNotFoundException dnfx)
            {
                Console.WriteLine("指定されたディレクトリが見つかりませんでした。");
                Console.WriteLine(dnfx.Message);
            }
            catch (IOException iox)
            {
                Console.WriteLine(iox.Message);
            }
            catch (UnauthorizedAccessException uax)
            {
                Console.WriteLine("ディレクトリにアクセスする権限がありません。");
                Console.WriteLine(uax.Message);
            }
            catch (ArgumentException ax)
            {
                Console.WriteLine("指定されたパスが正しくありません。");
                Console.WriteLine(ax.Message);
            }
            finally
            {
                if (testDirectoriesMade)
                {
                    CleanupTestDirectories(directoriesToSearch);
                }
            }
        }

        private static void ShowUsage()
        {
            Console.WriteLine("重複ファイル検索");
            Console.WriteLine("================");
            Console.WriteLine(
                "1つ以上のディレクトリに重複ファイルがあるかどうかをチェックします。");
            Console.WriteLine();
            Console.WriteLine(
                "使用方法: findduplicatefiles [/sub] ディレクトリ名 [ディレクトリ名] ...");
            Console.WriteLine("/sub - サブディレクトリ以下を再帰的に探索します");
            Console.ReadKey();
        }

        private static List<FileNameGroup> InspectDirectories(
            bool recurseIntoSubdirectories,
            IEnumerable<string> directoriesToSearch)
        {
            var searchOption = recurseIntoSubdirectories ?
                SearchOption.AllDirectories : SearchOption.TopDirectoryOnly;

            // 検索対象の全てのディレクトリに含まれるすべてのファイルパスを取得します

            var allFilePaths = from directory in directoriesToSearch
                               from file in GetDirectoryFiles(directory,
                                                              searchOption)
                               select file;

            // ローカルファイル名(親ディレクトリのパスを含まないファイル名)で
            // ファイルをグループ化しつつ、同名のローカルファイル名を持つ
            // 各ファイルの詳細を含んだリストを作成します。

            var fileNameGroups = from filePath in allFilePaths
                                 let fileNameWithoutPath = Path.GetFileName(filePath)
                                 group filePath by fileNameWithoutPath into nameGroup
                                 select new FileNameGroup
                                 {
                                     FileNameWithoutPath = nameGroup.Key,
                                     FilesWithThisName = GetDetails(nameGroup).ToList()
                                 };

            return fileNameGroups.ToList();
        }

        // 例11-33 内容を比較するようにDisplayMatchesを修正
        private static void DisplayMatches(
            IEnumerable<FileNameGroup> filesGroupedByName)
        {
            var groupsWithMoreThanOneFile = from nameGroup in filesGroupedByName
                                            where nameGroup.FilesWithThisName.Count > 1
                                            select nameGroup;

            foreach (var fileNameGroup in groupsWithMoreThanOneFile)
            {
                // 見つかったファイルをサイズでグループ化して、
                // 同じサイズのファイルが1つ以上ある場合には選択します
                var matchesBySize = from file in fileNameGroup.FilesWithThisName
                                    group file by file.FileSize into sizeGroup
                                    where sizeGroup.Count() > 1
                                    select sizeGroup;

                foreach (var matchedBySize in matchesBySize)
                {
                    string fileNameAndSize = string.Format("{0} ({1} bytes)",
                    fileNameGroup.FileNameWithoutPath, matchedBySize.Key);
                    WriteWithUnderlines(fileNameAndSize);

                    // 現在のファイルを含む各ディレクトリを表示します
                    {
                        List<FileContents> content = LoadFiles(matchedBySize);
                        CompareFiles(content);
                    }
                    Console.WriteLine();
                }
            }
        }

        // 例11-35 バイナリファイルの内容の読み込み
        private static List<FileContents> LoadFiles(IEnumerable<FileDetails> fileList)
        {
            var content = new List<FileContents>();
            foreach (FileDetails item in fileList)
            {
                byte[] contents = File.ReadAllBytes(item.FilePath);
                content.Add(new FileContents
                {
                    FilePath = item.FilePath,
                    Content = contents
                });
            }
            return content;
        }

        // 例11-36 CompareFilesメソッド
        private static void CompareFiles(List<FileContents> files)
        {
            Dictionary<FileContents, List<FileContents>> potentiallyMatched =
            BuildPotentialMatches(files);

            // この位置ですべてのbyteをそれぞれ比較します
            CompareBytes(files, potentiallyMatched);

            DisplayResults(files, potentiallyMatched);
        }

        // 例11-37 重複する可能性のある組み合わせを生成
        private static Dictionary<FileContents, List<FileContents>>
            BuildPotentialMatches(List<FileContents> files)
        {
            // 項目が以下のようになるディクショナリを生成します
            // { 0, { 1, 2, 3, 4, ... N } }
            // { 1, { 2, 3, 4, ... N }
            // ...
            // { N - 1, { N } }
            // ここでNはファイル数より１少ない数です
            var allCombinations = Enumerable.Range(0, files.Count - 1).ToDictionary(
                x => files[x],
            x => files.Skip(x + 1).ToList());

            return allCombinations;
        }

        // 例11-38 重複ファイルの表示
        private static void DisplayResults(
            List<FileContents> files,
            Dictionary<FileContents, List<FileContents>> currentlyMatched)
        {
            if (currentlyMatched.Count == 0) { return; }

            var alreadyMatched = new List<FileContents>();

            Console.WriteLine("重複");

            foreach (var matched in currentlyMatched)
            {
                // すでに重複している場合は無視します。
                if (alreadyMatched.Contains(matched.Key))
                {
                    continue;
                }
                else
                {
                    alreadyMatched.Add(matched.Key);
                }
                Console.WriteLine("-------");
                Console.WriteLine(matched.Key.FilePath);
                foreach (var file in matched.Value)
                {
                    Console.WriteLine(file.FilePath);
                    alreadyMatched.Add(file);
                }
            }
            Console.WriteLine("-------");
        }

        // 例11-39 重複の可能性があるすべてのファイル同士をbyte単位で比較
        private static void CompareBytes(
            List<FileContents> files,
            Dictionary<FileContents, List<FileContents>> potentiallyMatched)
        {
            // 同じサイズのファイル同士だけが対象となることに注意してください
            int fileLength = files[0].Content.Length;
            var sourceFilesWithNoMatches = new List<FileContents>();
            for (int fileByteOffset = 0; fileByteOffset < fileLength; ++fileByteOffset)
            {
                foreach (var sourceFileEntry in potentiallyMatched)
                {
                    byte[] sourceContent = sourceFileEntry.Key.Content;
                    for (int otherIndex = 0; otherIndex < sourceFileEntry.Value.Count;
                                                                            ++otherIndex)
                    {
                        // 2つのファイルの各i番目のバイトをチェックして、
                        // 一致しなければそれらをコレクションから削除
                        byte[] otherContent =
                            sourceFileEntry.Value[otherIndex].Content;
                        if (sourceContent[fileByteOffset] != otherContent[fileByteOffset])
                        {
                            sourceFileEntry.Value.RemoveAt(otherIndex);
                            otherIndex -= 1;
                            if (sourceFileEntry.Value.Count == 0)
                            {
                                sourceFilesWithNoMatches.Add(sourceFileEntry.Key);
                            }
                        }
                    }
                }
                foreach (FileContents fileWithNoMatches in sourceFilesWithNoMatches)
                {
                    potentiallyMatched.Remove(fileWithNoMatches);
                }
                // 重複の可能性があるファイルがこれ以上ない場合は、
                // 残りのファイルの処理を省略します
                if (potentiallyMatched.Count == 0)
                {
                    break;
                }
                sourceFilesWithNoMatches.Clear();
            }
        }

        private static void WriteWithUnderlines(string text)
        {
            Console.WriteLine(text);
            Console.WriteLine(new string('-', text.Length));
        }

        private static string[] MakeTestDirectories()
        {
            string localApplicationData = Path.Combine(
                Environment.GetFolderPath(
                    Environment.SpecialFolder.LocalApplicationData),
                @"Programming CSharp\FindDuplicates");

            // ログイン名を取得します。
            string userName = WindowsIdentity.GetCurrent().Name;
            // アクセスコントロールルールを作成します。
            FileSystemAccessRule fsarAllow =
                new FileSystemAccessRule(
                    userName,
                    FileSystemRights.FullControl,
                    AccessControlType.Allow);
            DirectorySecurity ds = new DirectorySecurity();
            ds.AddAccessRule(fsarAllow);

            FileSystemAccessRule fsarDeny =
                new FileSystemAccessRule(
                    userName,
                    FileSystemRights.WriteExtendedAttributes,
                    AccessControlType.Deny);
            ds.AddAccessRule(fsarDeny);

            // 3つのテスト用ディレクトリを作成します。
            // そしてアクセスが拒否された場合のテストを行うために
            // 4つめのディレクトリを作成できるようにします。
            var directories = new string[4];
            for (int i = 0; i < directories.Length - 1; ++i)
            {
                string directory = Path.GetRandomFileName();
                // 新しいランダムなファイル/ディレクトリ名と
                // LocalApplicationDataのフォルダ名を連結します。
                string fullPath = Path.Combine(localApplicationData, directory);
                // そしてディレクトリを作成します。
                Directory.CreateDirectory(fullPath, ds);
                directories[i] = fullPath;
                Console.WriteLine(fullPath);
            }

            CreateTestFiles(directories.Take(3));

            directories[3] = CreateDeniedDirectory(localApplicationData);

            return directories;
        }

        private static void CleanupTestDirectories(IEnumerable<string> directories)
        {
            foreach (var directory in directories)
            {
                AllowAccess(directory);
                Directory.Delete(directory, true);
            }
        }

        private static void CreateTestFiles(IEnumerable<string> directories)
        {
            string fileForAllDirectories = "SameNameAndContent.txt";
            string fileSameInAllButDifferentSizes = "SameNameDifferentSize.txt";
            string fileSameSizeInAllButDifferentContent =
                "SameNameAndSizeDifferentContent.txt";

            int directoryIndex = 0;
            // 各ディレクトリ内に個別のファイルを作成します。
            foreach (string directory in directories)
            {
                directoryIndex++;

                // このディレクトリにしかないファイルを作成します。
                string filename = Path.GetRandomFileName();
                string fullPath = Path.Combine(directory, filename);
                CreateFile(fullPath, "中身の例 1");

                // 次に中身が同じファイルをすべてのディレクトリで作成します。
                fullPath = Path.Combine(directory, fileForAllDirectories);
                CreateFile(fullPath, "すべてのディレクトリで見つかります");

                // 次に同じ名前ですがサイズが異なるファイルを
                // すべてのディレクトリで作成します。
                fullPath = Path.Combine(directory, fileSameInAllButDifferentSizes);

                StringBuilder builder = new StringBuilder();
                builder.AppendLine("こんどは");
                builder.AppendLine(new string('x', directoryIndex));
                CreateFile(fullPath, builder.ToString());

                // 同じ長さで、中身の異なるファイルを作成します
                fullPath = Path.Combine(directory, fileSameSizeInAllButDifferentContent);

                builder = new StringBuilder();
                builder.Append("こんどは");
                builder.Append(directoryIndex);
                builder.AppendLine("追加");
                CreateFile(fullPath, builder.ToString());
            }
        }

        private static void CreateFile(string fullPath, string contents)
        {
            File.WriteAllText(fullPath, contents);
        }

        static void DoubtfulCode()
        {
            if (File.Exists("SomeFile.txt"))
            {
                // ファイルを処理します
            }
        }

        private static string CreateDeniedDirectory(string parentPath)
        {
            string deniedDirectory = Path.GetRandomFileName();
            string fullDeniedPath = Path.Combine(parentPath, deniedDirectory);
            string userName = WindowsIdentity.GetCurrent().Name;
            DirectorySecurity ds = new DirectorySecurity();
            FileSystemAccessRule fsarDeny =
                new FileSystemAccessRule(
                    userName,
                    FileSystemRights.ListDirectory,
                    AccessControlType.Deny);
            ds.AddAccessRule(fsarDeny);

            Directory.CreateDirectory(fullDeniedPath, ds);
            return fullDeniedPath;
        }

        private static void AllowAccess(string directory)
        {
            DirectorySecurity ds = Directory.GetAccessControl(directory);

            string userName = WindowsIdentity.GetCurrent().Name;

            // 拒否ルールを削除
            FileSystemAccessRule fsarDeny =
                new FileSystemAccessRule(
                    userName,
                    FileSystemRights.ListDirectory,
                    AccessControlType.Deny);
            ds.RemoveAccessRuleSpecific(fsarDeny);

            // 許可ルールを追加
            FileSystemAccessRule fsarAllow =
                new FileSystemAccessRule(
                    userName,
                    FileSystemRights.FullControl,
                    AccessControlType.Allow);
            ds.AddAccessRule(fsarAllow);

            Directory.SetAccessControl(directory, ds);
        }

        private static IEnumerable<string> GetDirectoryFiles(
            string directory, SearchOption searchOption)
        {
            try
            {
                return Directory.GetFiles(directory, "*.*", searchOption);
            }
            catch (DirectoryNotFoundException dnfx)
            {
                Console.WriteLine("警告: 指定されたディレクトリが見つかりません。");
                Console.WriteLine(dnfx.Message);
            }
            catch (UnauthorizedAccessException uax)
            {
                Console.WriteLine(
                    "警告: このディレクトリにアクセスする権限がありません。");
                Console.WriteLine(uax.Message);
            }

            return Enumerable.Empty<string>();
        }

        private static IEnumerable<FileDetails> GetDetails(IEnumerable<string> paths)
        {
            foreach (string filePath in paths)
            {
                FileDetails details = null;
                try
                {
                    FileInfo info = new FileInfo(filePath);
                    details = new FileDetails
                    {
                        FilePath = filePath,
                        FileSize = info.Length
                    };
                }
                catch (FileNotFoundException fnfx)
                {
                    Console.WriteLine("警告: 指定されたファイルが見つかりません。");
                    Console.WriteLine(fnfx.Message);
                }
                catch (IOException iox)
                {
                    Console.Write("警告: ");
                    Console.WriteLine(iox.Message);
                }
                catch (UnauthorizedAccessException uax)
                {
                    Console.WriteLine(
                        "警告: このファイルにアクセスするパーミッションがありません。");
                    Console.WriteLine(uax.Message);
                }
                if (details != null)
                {
                    yield return details;
                }
            }
        }
    }
}
