﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Security.AccessControl;
using System.Security.Principal;
using System.Text;

namespace DuplicateFinder
{
    class Program
    {
        static void Main(string[] args)
        {
            bool recurseIntoSubdirectories = false;

            if (args.Length < 1)
            {
                ShowUsage();
                return;
            }

            int firstDirectoryIndex = 0;
            IEnumerable<string> directoriesToSearch = null;
            bool testDirectoriesMade = false;

            // テストモードかどうかチェック
            if (args.Length == 1 && args[0] == "/test")
            {
                directoriesToSearch = MakeTestDirectories();
                testDirectoriesMade = true;
                recurseIntoSubdirectories = true;
            }
            else
            {
                if (args.Length > 1)
                {
                    // 再帰処理が必要かどうかチェック
                    if (args[0] == "/sub")
                    {
                        if (args.Length < 2)
                        {
                            ShowUsage();
                            return;
                        }
                        recurseIntoSubdirectories = true;
                        firstDirectoryIndex = 1;
                    }
                }

                // コマンドラインからディレクトリのリストを取得
                directoriesToSearch = args.Skip(firstDirectoryIndex);
            }

            try
            {
                List<FileNameGroup> filesGroupedByName =
                    InspectDirectories(recurseIntoSubdirectories, directoriesToSearch);

                DisplayMatches(filesGroupedByName);

                Console.ReadKey();
            }
            catch (PathTooLongException ptlx)
            {
                Console.WriteLine("指定されたパスが長すぎます。");
                Console.WriteLine(ptlx.Message);
            }
            catch (DirectoryNotFoundException dnfx)
            {
                Console.WriteLine("指定されたディレクトリが見つかりませんでした。");
                Console.WriteLine(dnfx.Message);
            }
            catch (IOException iox)
            {
                Console.WriteLine(iox.Message);
            }
            catch (UnauthorizedAccessException uax)
            {
                Console.WriteLine("ディレクトリにアクセスする権限がありません。");
                Console.WriteLine(uax.Message);
            }
            catch (ArgumentException ax)
            {
                Console.WriteLine("指定されたパスが正しくありません。");
                Console.WriteLine(ax.Message);
            }
            finally
            {
                if (testDirectoriesMade)
                {
                    CleanupTestDirectories(directoriesToSearch);
                }
            }
        }

        private static void ShowUsage()
        {
            Console.WriteLine("重複ファイル検索");
            Console.WriteLine("================");
            Console.WriteLine(
                "1つ以上のディレクトリに重複ファイルがあるかどうかをチェックします。");
            Console.WriteLine();
            Console.WriteLine(
                "使用方法: findduplicatefiles [/sub] ディレクトリ名 [ディレクトリ名] ...");
            Console.WriteLine("/sub - サブディレクトリ以下を再帰的に探索します");
            Console.ReadKey();
        }

        private static List<FileNameGroup> InspectDirectories(
            bool recurseIntoSubdirectories,
            IEnumerable<string> directoriesToSearch)
        {
            var searchOption = recurseIntoSubdirectories ?
                SearchOption.AllDirectories : SearchOption.TopDirectoryOnly;

            // 検索対象の全てのディレクトリに含まれるすべてのファイルパスを取得します

            // 例11-30 エラーを想定した走査
            var allFilePaths = from directory in directoriesToSearch
                               from file in GetDirectoryFiles(directory,
                                                              searchOption)
                               select file;

            // ローカルファイル名(親ディレクトリのパスを含まないファイル名)で
            // ファイルをグループ化しつつ、同名のローカルファイル名を持つ
            // 各ファイルの詳細を含んだリストを作成します。

            // 例11-32 エラーを処理しつつ詳細を取得する
            var fileNameGroups = from filePath in allFilePaths
                                 let fileNameWithoutPath = Path.GetFileName(filePath)
                                 group filePath by fileNameWithoutPath into nameGroup
                                 select new FileNameGroup
                                 {
                                     FileNameWithoutPath = nameGroup.Key,
                                     FilesWithThisName = GetDetails(nameGroup).ToList()
                                 };

            return fileNameGroups.ToList();
        }

        private static void DisplayMatches(
            IEnumerable<FileNameGroup> filesGroupedByName)
        {
            var groupsWithMoreThanOneFile = from nameGroup in filesGroupedByName
                                            where nameGroup.FilesWithThisName.Count > 1
                                            select nameGroup;

            foreach (var fileNameGroup in groupsWithMoreThanOneFile)
            {
                // 見つかったファイルをサイズでグループ化して、
                // 同じサイズのファイルが1つ以上ある場合には選択します
                var matchesBySize = from file in fileNameGroup.FilesWithThisName
                                    group file by file.FileSize into sizeGroup
                                    where sizeGroup.Count() > 1
                                    select sizeGroup;

                foreach (var matchedBySize in matchesBySize)
                {
                    string fileNameAndSize = string.Format("{0} ({1} bytes)",
                    fileNameGroup.FileNameWithoutPath, matchedBySize.Key);
                    WriteWithUnderlines(fileNameAndSize);
                    // 重複ファイルを含むそれぞれのディレクトリを表示
                    foreach (var file in matchedBySize)
                    {
                        Console.WriteLine(Path.GetDirectoryName(file.FilePath));
                    }
                    Console.WriteLine();
                }
            }
        }

        private static void WriteWithUnderlines(string text)
        {
            Console.WriteLine(text);
            Console.WriteLine(new string('-', text.Length));
        }

        private static string[] MakeTestDirectories()
        {
            string localApplicationData = Path.Combine(
                Environment.GetFolderPath(
                    Environment.SpecialFolder.LocalApplicationData),
                @"Programming CSharp\FindDuplicates");

            // ログイン名を取得します。
            string userName = WindowsIdentity.GetCurrent().Name;
            // アクセスコントロールルールを作成します。
            FileSystemAccessRule fsarAllow =
                new FileSystemAccessRule(
                    userName,
                    FileSystemRights.FullControl,
                    AccessControlType.Allow);
            DirectorySecurity ds = new DirectorySecurity();
            ds.AddAccessRule(fsarAllow);

            FileSystemAccessRule fsarDeny =
                new FileSystemAccessRule(
                    userName,
                    FileSystemRights.WriteExtendedAttributes,
                    AccessControlType.Deny);
            ds.AddAccessRule(fsarDeny);

            // 3つのテスト用ディレクトリを作成します。
            // そしてアクセスが拒否された場合のテストを行うために
            // 4つめのディレクトリを作成できるようにします。
            var directories = new string[4];
            for (int i = 0; i < directories.Length - 1; ++i)
            {
                string directory = Path.GetRandomFileName();
                // 新しいランダムなファイル/ディレクトリ名と
                // LocalApplicationDataのフォルダ名を連結します。
                string fullPath = Path.Combine(localApplicationData, directory);
                // そしてディレクトリを作成します。
                Directory.CreateDirectory(fullPath, ds);
                directories[i] = fullPath;
                Console.WriteLine(fullPath);
            }

            CreateTestFiles(directories.Take(3));

            directories[3] = CreateDeniedDirectory(localApplicationData);

            return directories;
        }

        private static void CleanupTestDirectories(IEnumerable<string> directories)
        {
            foreach (var directory in directories)
            {
                AllowAccess(directory);
                Directory.Delete(directory, true);
            }
        }

        private static void CreateTestFiles(IEnumerable<string> directories)
        {
            string fileForAllDirectories = "SameNameAndContent.txt";
            string fileSameInAllButDifferentSizes = "SameNameDifferentSize.txt";
            int directoryIndex = 0;
            // 各ディレクトリ内に個別のファイルを作成します。
            foreach (string directory in directories)
            {
                directoryIndex++;

                // このディレクトリにしかないファイルを作成します。
                string filename = Path.GetRandomFileName();
                string fullPath = Path.Combine(directory, filename);
                CreateFile(fullPath, "中身の例 1");

                // 次に中身が同じファイルをすべてのディレクトリで作成します。
                fullPath = Path.Combine(directory, fileForAllDirectories);
                CreateFile(fullPath, "すべてのディレクトリで見つかります");

                // 次に同じ名前ですがサイズが異なるファイルを
                // すべてのディレクトリで作成します。
                fullPath = Path.Combine(directory, fileSameInAllButDifferentSizes);

                StringBuilder builder = new StringBuilder();
                builder.AppendLine("こんどは");
                builder.AppendLine(new string('x', directoryIndex));
                CreateFile(fullPath, builder.ToString());
            }
        }

        private static void CreateFile(string fullPath, string contents)
        {
            File.WriteAllText(fullPath, contents);
        }

        static void DoubtfulCode()
        {
            if (File.Exists("SomeFile.txt"))
            {
                // ファイルを処理します
            }
        }

        private static string CreateDeniedDirectory(string parentPath)
        {
            string deniedDirectory = Path.GetRandomFileName();
            string fullDeniedPath = Path.Combine(parentPath, deniedDirectory);
            string userName = WindowsIdentity.GetCurrent().Name;
            DirectorySecurity ds = new DirectorySecurity();
            FileSystemAccessRule fsarDeny =
                new FileSystemAccessRule(
                    userName,
                    FileSystemRights.ListDirectory,
                    AccessControlType.Deny);
            ds.AddAccessRule(fsarDeny);

            Directory.CreateDirectory(fullDeniedPath, ds);
            return fullDeniedPath;
        }

        private static void AllowAccess(string directory)
        {
            DirectorySecurity ds = Directory.GetAccessControl(directory);

            string userName = WindowsIdentity.GetCurrent().Name;

            // 拒否ルールを削除
            FileSystemAccessRule fsarDeny =
                new FileSystemAccessRule(
                    userName,
                    FileSystemRights.ListDirectory,
                    AccessControlType.Deny);
            ds.RemoveAccessRuleSpecific(fsarDeny);

            // 許可ルールを追加
            FileSystemAccessRule fsarAllow =
                new FileSystemAccessRule(
                    userName,
                    FileSystemRights.FullControl,
                    AccessControlType.Allow);
            ds.AddAccessRule(fsarAllow);

            Directory.SetAccessControl(directory, ds);
        }

        // 例11-29 ヘルパーメソッドとして例外処理を行う
        private static IEnumerable<string> GetDirectoryFiles(
            string directory, SearchOption searchOption)
        {
            try
            {
                return Directory.GetFiles(directory, "*.*", searchOption);
            }
            catch (DirectoryNotFoundException dnfx)
            {
                Console.WriteLine("警告: 指定されたディレクトリが見つかりません。");
                Console.WriteLine(dnfx.Message);
            }
            catch (UnauthorizedAccessException uax)
            {
                Console.WriteLine(
                    "警告: このディレクトリにアクセスする権限がありません。");
                Console.WriteLine(uax.Message);
            }

            return Enumerable.Empty<string>();
        }

        // 例11-31 FileInfoからのエラーを処理
        private static IEnumerable<FileDetails> GetDetails(IEnumerable<string> paths)
        {
            foreach (string filePath in paths)
            {
                FileDetails details = null;
                try
                {
                    FileInfo info = new FileInfo(filePath);
                    details = new FileDetails
                    {
                        FilePath = filePath,
                        FileSize = info.Length
                    };
                }
                catch (FileNotFoundException fnfx)
                {
                    Console.WriteLine("警告: 指定されたファイルが見つかりません。");
                    Console.WriteLine(fnfx.Message);
                }
                catch (IOException iox)
                {
                    Console.Write("警告: ");
                    Console.WriteLine(iox.Message);
                }
                catch (UnauthorizedAccessException uax)
                {
                    Console.WriteLine(
                        "警告: このファイルにアクセスするパーミッションがありません。");
                    Console.WriteLine(uax.Message);
                }
                if (details != null)
                {
                    yield return details;
                }
            }
        }
    }
}
