﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Security.AccessControl;
using System.Security.Principal;
using System.Text;

namespace DuplicateFinder
{
    class Program
    {
        static void Main(string[] args)
        {
            bool recurseIntoSubdirectories = false;

            if (args.Length < 1)
            {
                ShowUsage();
                return;
            }

            int firstDirectoryIndex = 0;
            IEnumerable<string> directoriesToSearch = null;
            bool testDirectoriesMade = false;

            try
            {
                // テストモードかどうかチェック
                if (args.Length == 1 && args[0] == "/test")
                {
                    directoriesToSearch = MakeTestDirectories();
                    testDirectoriesMade = true;
                    recurseIntoSubdirectories = true;
                }
                else
                {
                    if (args.Length > 1)
                    {
                        // 再帰処理が必要かどうかチェック
                        if (args[0] == "/sub")
                        {
                            if (args.Length < 2)
                            {
                                ShowUsage();
                                return;
                            }
                            recurseIntoSubdirectories = true;
                            firstDirectoryIndex = 1;
                        }
                    }

                    // コマンドラインからディレクトリのリストを取得
                    directoriesToSearch = args.Skip(firstDirectoryIndex);
                }

                List<FileNameGroup> filesGroupedByName =
                    InspectDirectories(recurseIntoSubdirectories, directoriesToSearch);

                DisplayMatches(filesGroupedByName);

                Console.ReadKey();
            }
            finally
            {
                if (testDirectoriesMade)
                {
                    CleanupTestDirectories(directoriesToSearch);
                }
            }
        }

        private static void ShowUsage()
        {
            Console.WriteLine("重複ファイル検索");
            Console.WriteLine("================");
            Console.WriteLine(
                "1つ以上のディレクトリに重複ファイルがあるかどうかをチェックします。");
            Console.WriteLine();
            Console.WriteLine(
                "使用方法: findduplicatefiles [/sub] ディレクトリ名 [ディレクトリ名] ...");
            Console.WriteLine("/sub - サブディレクトリ以下を再帰的に探索します");
            Console.ReadKey();
        }

        private static List<FileNameGroup> InspectDirectories(
            bool recurseIntoSubdirectories,
            IEnumerable<string> directoriesToSearch)
        {
            var searchOption = recurseIntoSubdirectories ?
                SearchOption.AllDirectories : SearchOption.TopDirectoryOnly;

            // 検索対象の全てのディレクトリに含まれるすべてのファイルパスを取得します
            var allFilePaths = from directory in directoriesToSearch
                               from file in Directory.GetFiles(directory, "*.*",
                                                                searchOption)
                               select file;

            // ローカルファイル名(親ディレクトリのパスを含まないファイル名)で
            // ファイルをグループ化しつつ、同名のローカルファイル名を持つ
            // 各ファイルの詳細を含んだリストを作成します。

            var fileNameGroups = from filePath in allFilePaths
                                 let fileNameWithoutPath = Path.GetFileName(filePath)
                                 group filePath by fileNameWithoutPath into nameGroup
                                 select new FileNameGroup
                                 {
                                     FileNameWithoutPath = nameGroup.Key,
                                     FilesWithThisName =
                                      (from filePath in nameGroup
                                       let info = new FileInfo(filePath)
                                       select new FileDetails
                                       {
                                           FilePath = filePath,
                                           FileSize = info.Length
                                       }).ToList()
                                 };

            return fileNameGroups.ToList();
        }

        private static void DisplayMatches(
            IEnumerable<FileNameGroup> filesGroupedByName)
        {
            var groupsWithMoreThanOneFile = from nameGroup in filesGroupedByName
                                            where nameGroup.FilesWithThisName.Count > 1
                                            select nameGroup;

            foreach (var fileNameGroup in groupsWithMoreThanOneFile)
            {
                // 見つかったファイルをサイズでグループ化して、
                // 同じサイズのファイルが1つ以上ある場合には選択します
                var matchesBySize = from file in fileNameGroup.FilesWithThisName
                                    group file by file.FileSize into sizeGroup
                                    where sizeGroup.Count() > 1
                                    select sizeGroup;

                foreach (var matchedBySize in matchesBySize)
                {
                    string fileNameAndSize = string.Format("{0} ({1} bytes)",
                    fileNameGroup.FileNameWithoutPath, matchedBySize.Key);
                    WriteWithUnderlines(fileNameAndSize);
                    // 重複ファイルを含むそれぞれのディレクトリを表示
                    foreach (var file in matchedBySize)
                    {
                        Console.WriteLine(Path.GetDirectoryName(file.FilePath));
                    }
                    Console.WriteLine();
                }
            }
        }

        private static void WriteWithUnderlines(string text)
        {
            Console.WriteLine(text);
            Console.WriteLine(new string('-', text.Length));
        }

        // 例11-15 新しいディレクトリへのアクセスコントロールを設定する
        private static string[] MakeTestDirectories()
        {
            string localApplicationData = Path.Combine(
                Environment.GetFolderPath(
                    Environment.SpecialFolder.LocalApplicationData),
                @"Programming CSharp\FindDuplicates");

            // ログイン名を取得します。
            string userName = WindowsIdentity.GetCurrent().Name;
            // アクセスコントロールルールを作成します。
            FileSystemAccessRule fsarAllow =
                new FileSystemAccessRule(
                    userName,
                    FileSystemRights.FullControl,
                    AccessControlType.Allow);
            DirectorySecurity ds = new DirectorySecurity();
            ds.AddAccessRule(fsarAllow);

            // 例11-16 パーミッションの拒否
            FileSystemAccessRule fsarDeny =
                new FileSystemAccessRule(
                    userName,
                    FileSystemRights.WriteExtendedAttributes,
                    AccessControlType.Deny);
            ds.AddAccessRule(fsarDeny);

            // 3つのテスト用ディレクトリを作成します。
            var directories = new string[3];
            for (int i = 0; i < directories.Length; ++i)
            {
                string directory = Path.GetRandomFileName();
                // 新しいランダムなファイル/ディレクトリ名と
                // LocalApplicationDataのフォルダ名を連結します。
                string fullPath = Path.Combine(localApplicationData, directory);
                // そしてディレクトリを作成します。
                Directory.CreateDirectory(fullPath, ds);
                directories[i] = fullPath;
                Console.WriteLine(fullPath);
            }

            CreateTestFiles(directories);
            return directories;
        }

        // 例11-17 ディレクトリの削除
        private static void CleanupTestDirectories(IEnumerable<string> directories)
        {
            foreach (var directory in directories)
            {
                Directory.Delete(directory);
            }
        }

        // 例11-19 CreateTestFilesメソッド
        private static void CreateTestFiles(IEnumerable<string> directories)
        {
            string fileForAllDirectories = "SameNameAndContent.txt";
            string fileSameInAllButDifferentSizes = "SameNameDifferentSize.txt";
            int directoryIndex = 0;
            // 各ディレクトリ内に個別のファイルを作成します。
            foreach (string directory in directories)
            {
                directoryIndex++;

                // このディレクトリにしかないファイルを作成します。
                string filename = Path.GetRandomFileName();
                string fullPath = Path.Combine(directory, filename);
                CreateFile(fullPath, "中身の例 1");

                // 次に中身が同じファイルをすべてのディレクトリで作成します。
                fullPath = Path.Combine(directory, fileForAllDirectories);
                CreateFile(fullPath, "すべてのディレクトリで見つかります");

                // 次に同じ名前ですがサイズが異なるファイルを
                // すべてのディレクトリで作成します。
                fullPath = Path.Combine(directory, fileSameInAllButDifferentSizes);

                StringBuilder builder = new StringBuilder();
                builder.AppendLine("こんどは");
                builder.AppendLine(new string('x', directoryIndex));
                CreateFile(fullPath, builder.ToString());
            }
        }

        // 例11-20 文字列を新しいファイルに書き込む
        private static void CreateFile(string fullPath, string contents)
        {
            File.WriteAllText(fullPath, contents);
        }
    }
}
