﻿using System;
using System.Linq;
using System.ServiceModel;
using System.Collections.Generic;

namespace ChatServerLibrary
{
    [ServiceBehavior(
        InstanceContextMode = InstanceContextMode.Single,
        ConcurrencyMode = ConcurrencyMode.Reentrant)]
    public class ChatService : IChatService
    {
        private Dictionary<IChatClient, string> clientsAndNames =
            new Dictionary<IChatClient, string>();

        public bool Connect(string name)
        {
            if (clientsAndNames.ContainsValue(name))
            {
                // 既に同じ名前が使用中のため、接続を拒否します。
                return false;
            }

            IChatClient clientCallback =
                OperationContext.Current.GetCallbackChannel<IChatClient>();

            // clientsAndNamesは共有されますが、
            // ConcurrentMode.Reentrantを指定しているためメッセージが
            // 一度に1つしか処理されないことが保証されているはずなので
            // ここではロックを行っていません。
            clientsAndNames.Add(clientCallback, name);
            Console.WriteLine(name + "が接続しました");

            return true;
        }

        public void PostNote(string note)
        {
            IChatClient clientCallback =
                OperationContext.Current.GetCallbackChannel<IChatClient>();
            string name = clientsAndNames[clientCallback];

            Console.WriteLine("{0}: {1}", name, note);

            // ToArray()はコレクションのコピーを作成します。
            // そうすることによって、もしもループ中にクライアントの接続を
            // 切らなければならなくなったとしても、コレクションを変更して
            // しまうために例外が発生するような状況を回避できるようになります。
            KeyValuePair<IChatClient, string>[] copiedNames =
                clientsAndNames.ToArray();
            foreach (KeyValuePair<IChatClient, string> client in copiedNames)
            {
                // メッセージを送信してきたクライアントはどのようなノートが入力されたのか
                // 既にわかっているので、このクライアントに対しては返信しないようにします。
                if (client.Key != clientCallback)
                {
                    Console.WriteLine("{0} にノートを送信中", client.Value);
                    try
                    {
                        client.Key.NotePosted(name, note);
                    }
                    catch (Exception x)
                    {
                        Console.WriteLine("エラー: {0}", x);
                        DisconnectClient(client.Key);
                    }
                }
            }
        }

        public void Disconnect()
        {
            IChatClient clientCallback =
                OperationContext.Current.GetCallbackChannel<IChatClient>();
            DisconnectClient(clientCallback);
        }

        private void DisconnectClient(IChatClient clientCallback)
        {
            string name = clientsAndNames[clientCallback];
            Console.WriteLine(name + " の接続を切りました");
            clientsAndNames.Remove(clientCallback);
        }
    }
}
