﻿using System;
using System.Configuration;
using System.Data.EntityClient;
using System.Data.Objects;
using System.Linq;
using System.Transactions;

namespace EfExamples
{
    class Program
    {
        static void Main(string[] args)
        {
            // 例14-3 生成されたエンティティ型の利用
            using (var dbContext = new AdventureWorksLT2008Entities())
            {
                DateTime orderDate = new DateTime(2004, 6, 1);
                var orders = from order in dbContext.SalesOrderHeaders
                             where order.OrderDate == orderDate
                             select order;
                foreach (SalesOrderHeader order in orders)
                {
                    Console.WriteLine(order.TotalDue);
                }
            }

            // 例14-4 ナビゲーションプロパティの利用
            using (var dbContext = new AdventureWorksLT2008Entities())
            {
                var customerOrderCounts = from cust in dbContext.Customers
                                          select new
                                          {
                                              cust.CustomerID,
                                              OrderCount = cust.SalesOrderHeaders.Count
                                          };
                foreach (var customerInfo in customerOrderCounts)
                {
                    Console.WriteLine("顧客{0}は注文が{1}回あります",
                    customerInfo.CustomerID, customerInfo.OrderCount);
                }
            }

            using (var dbContext = new AdventureWorksLT2008Entities())
            {
                // 例14-5 ナビゲーションプロパティを利用して複数の関連性を利用する
                var info = from cust in dbContext.Customers
                           select new
                           {
                               cust.CustomerID,
                               Orders = from order in cust.SalesOrderHeaders
                                        where order.Status == 5
                                        select new
                                        {
                                            order.TotalDue,
                                            ItemCount = order.SalesOrderDetails.Count
                                        }
                           };

                foreach (var item in info)
                {
                    if (item.Orders.Count() > 0)
                    {
                        Console.WriteLine(item.CustomerID);
                        foreach (var order in item.Orders)
                        {
                            Console.WriteLine(" " + order);
                        }
                    }
                }

                {
                    // 例14-6 最初のクエリの実行後にアソシエーションを利用する
                    Customer myCustomer = dbContext.Customers.Single(
                        cust => cust.CustomerID == 29531);
                    Console.WriteLine(myCustomer.SalesOrderHeaders.Count);
                }

                {
                    // 例14-7 アソシエーションを利用するために明示的にエンティティをロードする
                    Customer myCustomer = dbContext.Customers.Single(
                    cust => cust.CustomerID == 29531);
                    myCustomer.SalesOrderHeaders.Load();
                    Console.WriteLine(myCustomer.SalesOrderHeaders.Count);
                }

                {
                    // 例14-8 関係を指定することにより事前ロードを行う
                    var customersWithOrderDetails = dbContext.Customers.
                        Include("SalesOrderHeaders.SalesOrderDetails");
                    Customer myCustomer = customersWithOrderDetails.Single(
                        cust => cust.CustomerID == 29531);
                    Console.WriteLine(myCustomer.SalesOrderHeaders.Count);
                }


                {
                    DateTime orderDate = new DateTime(2004, 6, 1);

                    // 例14-9 単純なLINQ to Entities クエリ式
                    var orders = from order in dbContext.SalesOrderHeaders
                                 where order.OrderDate == orderDate
                                 select order;

                    // この時点ではまだクエリは実行されません。
                    // クエリからデータを取得し始めようとすると、そこで初めて
                    // 実際の処理が行われます。

                    // 以下の例ではクエリ中で'orders'を参照しているにもかかわらず、
                    // 実際にはデータベースへのアクセスは行われません。
                    // 元々のクエリを拡張したバージョンのクエリを用意したにすぎないのです。

                    // 例14-10 クエリの連鎖
                    var orderedOrders = from order in orders
                                        orderby order.OrderDate
                                        select order;
                }

                {
                    DateTime orderDate = new DateTime(2004, 6, 1);

                    // 例14-11 同じ機能のクエリ
                    var orderedOrders = from order in dbContext.SalesOrderHeaders
                                        where order.OrderDate == orderDate
                                        orderby order.OrderDate
                                        select order;
                }
            }

            // 例14-12 ESQL によるクエリ
            using (var dbContext = new AdventureWorksLT2008Entities())
            {
                DateTime orderDate = new DateTime(2004, 6, 1);
                var query = dbContext.CreateQuery<SalesOrderHeader>("SELECT VALUE o " +
                        "FROM AdventureWorksLT2008Entities.SalesOrderHeaders AS o " +
                        "WHERE o.OrderDate = @orderDate",
                    new ObjectParameter("orderDate", orderDate));

                foreach (var order in query)
                {
                    Console.WriteLine(order.TotalDue);
                }
            }

            // 例14-14 明示的に接続文字列を指定する

            // ベースとなるデータプロバイダ用の接続文字列を取得します。
            ConnectionStringSettings dbConnectionInfo =
                ConfigurationManager.ConnectionStrings["AdventureWorksSql"];

            var csb = new EntityConnectionStringBuilder();
            csb.Provider = dbConnectionInfo.ProviderName;
            csb.ProviderConnectionString = dbConnectionInfo.ConnectionString;
            csb.Metadata = "res://*/AdventureWorksModel.csdl|" +
                "res://*/AdventureWorksModel.ssdl|res://*/AdventureWorksModel.msl";

            using (var dbContext = new AdventureWorksLT2008Entities(csb.ConnectionString))
            {
                foreach (SalesOrderHeader order in dbContext.SalesOrderHeaders)
                {
                    Console.WriteLine(order.TotalDue);
                }
            }


            // 例14-16 既存のエンティティを変更する
            using (var dbContext = new AdventureWorksLT2008Entities())
            {
                var orderQuery = from customer in dbContext.Customers
                                 where customer.CustomerID == 29531
                                 from order in customer.SalesOrderHeaders
                                 orderby order.OrderDate descending
                                 select order;
                SalesOrderHeader latestOrder = orderQuery.First();
                latestOrder.Comment = "発送準備が完了した時点で連絡を入れること";
                dbContext.SaveChanges();
            }

            try
            {
                using (var dbContext = new AdventureWorksLT2008Entities())
                {
                    // 例14-17 新しいエンティティの追加時に制約が満たされていない
                    SalesOrderDetail detail = new SalesOrderDetail();
                    dbContext.AddToSalesOrderDetails(detail);
                    // 以下のコードで例外がスローされます!
                    dbContext.SaveChanges();
                }
            }
            catch (Exception x)
            {
                Console.WriteLine(x);
            }

            using (var dbContext = new AdventureWorksLT2008Entities())
            {
                var orderQuery = from customer in dbContext.Customers
                                 where customer.CustomerID == 29531
                                 from order in customer.SalesOrderHeaders
                                 orderby order.OrderDate descending
                                 select order;
                SalesOrderHeader latestOrder = orderQuery.First();


                // 例14-18 新しいエンティティの追加
                // ...latestOrderは例14-16のようなコードを使用して取得した
                // SalesOrderHeaderオブジェクトです。
                SalesOrderDetail detail = new SalesOrderDetail();
                detail.SalesOrderHeader = latestOrder;
                detail.ModifiedDate = DateTime.Now;
                detail.OrderQty = 1;
                detail.ProductID = 680; // HLサイズのロードバイク用フレーム - 黒 58番
                dbContext.AddToSalesOrderDetails(detail);
                dbContext.SaveChanges();

                // 例14-19 エンティティの削除
                dbContext.DeleteObject(detail);
                dbContext.SaveChanges();
            }

            // 例14-20 TransactionScope
            using (var dbContext = new AdventureWorksLT2008Entities())
            {
                using (var txScope = new TransactionScope())
                {
                    var customersWithOrders = from cust in dbContext.Customers
                                              where cust.SalesOrderHeaders.Count > 0
                                              select cust;
                    
                    foreach (var customer in customersWithOrders)
                    {
                        Console.WriteLine("顧客{0}は注文が{1}回あります",
                        customer.CustomerID, customer.SalesOrderHeaders.Count);
                    }
                    
                    txScope.Complete();
                }
            }
        }
    }
}
